
// TALKING CLOCK using PIC16F876A, SD Card and 16X1 LCD module. By www.moty22.co.uk
//
// File will compile with Hi-Tech C compiler or XC8 on MPLAB or MPLABX.
// Recording is done by the ADC using only the LSB 8 bits. Bytes are written to the SD
// at the rate of 20KHz.   
// For playback CCP is used as a DAC (Digital to Analogue Converter).
   
#include <htc.h>

#if defined __XC8
  #pragma config LVP=OFF, FOSC=HS, CP=OFF, CPD=OFF, WDTE=OFF
#else
  __CONFIG(LVPDIS & UNPROTECT & WDTDIS & HS & UNPROTECT);
#endif

#define _XTAL_FREQ 8000000
#define __delay_us(x) _delay((unsigned long)((x)*(_XTAL_FREQ/4000000.0)))
#define __delay_ms(x) _delay((unsigned long)((x)*(_XTAL_FREQ/4000.0)))
#define CS RC6		//SD chip select input
#define RecLED RC7
#define menu RB0	//menu pushbutton
#define talk RB1	//play, talk, record PB
#define plus RB2	//increase PB
#define minus RB3	//decrease PB
#define rs RB4
#define e RB5
#define bk digit(32)	//blank character

//prototypes
void main(void);
void delay(unsigned char msec);
void initLCD(void);
void initSD(void);
void clock(void);
void digit(unsigned char fig);
void line(unsigned char msg);
unsigned char SPI(unsigned char data);
char Command(unsigned char frame1, unsigned long adrs, unsigned char frame2 );
void readSD(unsigned char mem);
void writeSD(unsigned char mem);

unsigned char min,sec,hour,m_trk,h_trk,gap;//
 
void main(void){
	
	// PIC I/O init
	TRISA = 0b11110000;
	TRISC = 0b10011;		// sck rc3, sdo rc5, sdi rc4, rc2 CPP1.
	TRISB = 0b1111;   	// switches
	 // pullup
    #if defined __XC8
    nRBPU = 0;			//RB pullup on
    #else
    RBPU = 0;			//RB pullup on
    #endif
	
	T1CON = 0b11111;        //TMR1 prescale 1:2, osc=on, timer on 0b110001
	RecLED = 0;
		
	//analogue init
	CCP1CON = 0B1100;	//PWM mode
	PR2 = 100;		//20KHz
	T2CON = 0B100;	//prescale 1, post scale 1, timer2 on
	ADCON1 = 0B11000010;		// Fosc/16, all analogue, right just. 
	ADCON0 = 0B1100000;		// Fosc/16. ch4, ADON=off
	PSA = T0CS = PS2 = 0;			//timer0 2MHz / 16 = 2 msec
	
	//SPI init
	SSPCON = 0B110010;	//low speed osc/64(125kHz),enabled,clock idle=H
	CS = 1; 		// disable SD	

	initLCD();
	initSD();
	
	min=0; sec=0; hour=0; m_trk=0; h_trk=0; gap=0;
	
	while(1) {
		line(1);
		//CLOCK
		while(1) {
			while(!TMR1IF){				//TMR1IF is set every 4 seconds
				if(!menu){break;}
				if(!talk){gap=1;}
			}
			if(!menu){break;}
			TMR1IF=0;
			++sec;	
	
//			if(sec > 14){
				sec=0; ++min;
					if(min > 59){
						min=0; ++hour;
							if(hour > 23){
								hour=0;
							}
					}
				line(1);	
//			}					
			
			if(gap){gap=0;readSD(hour+60); readSD(min);} 
		}	
			//MENU ITEMS
			while(!menu) {delay(50);}	//record minutes tracks
			line(4);
			while(menu){
				if(!plus){
					++m_trk;
					if(m_trk > 59) {m_trk=0;}
					line(4);
					delay(200);
				}
				if(!minus){	
					--m_trk;
					if(m_trk == 255) {m_trk=59;}
					line(4); 
					delay(200);
				}
				if(!talk){line(7); writeSD(m_trk); line(4);}	
			}
			
			while(!menu) {delay(50);}	//record hour tracks
			line(8);
			while(menu){
				if(!plus){
					++h_trk;
					if(h_trk > 23) {h_trk=0;}
					line(8);
					delay(200);
				}
				if(!minus){	
					--h_trk;
					if(h_trk == 255) {h_trk=23;}
					line(8); 
					delay(200);
				}
				if(!talk){line(7); writeSD(h_trk+60); line(8);}	
			}
							
			while(!menu) {delay(50);}	//play minutes tracks			
			line(6);
			while(menu){
				if(!plus){
					++m_trk;
					if(m_trk > 59) {m_trk=0;}
					line(6);
					delay(200);
				}
				if(!minus){	
					--m_trk;
					if(m_trk == 255) {m_trk=59;}
					line(6);
					delay(200);
				}
				if(!talk){readSD(m_trk);}	
			}
			
			while(!menu) {delay(50);}	//play hours tracks			
			line(9);
			while(menu){
				if(!plus){
					++h_trk;
					if(h_trk > 23) {h_trk=0;}
					line(9);
					delay(200);
				}
				if(!minus){	
					--h_trk;
					if(h_trk == 255) {h_trk=23;}
					line(9);
					delay(200);
				}
				if(!talk){readSD(h_trk+60);}	
			}				
			
			while(!menu){delay(50);}	//set hours
			line(2);
			while(menu){
				if(!plus){
					++hour;
					if(hour > 23) {hour=0;}
					line(2);
					delay(200);
				}
				if(!minus){	
					--hour;
					if(hour==255) {hour=23;}
					line(2);
					delay(200);
				}	
			}
			
			while(!menu) {delay(50);}	//set minutes
			line(3);
			while(menu){
				if(!plus){
					++min;
					if(min > 59) {min=0;}
					line(3);
					delay(200);
				}
				if(!minus){	
					--min;
					if(min == 255) {min=59;}
					line(3);
					delay(200);
				}	
			}
			while(!menu) {delay(50);}	
	}
	
}

void delay(unsigned char msec){		//units of 2 msec delay 
	
	while (msec){
		TMR0=0; TMR0IF=0;
		while(!TMR0IF){}
		--msec;
	}
}		

void initLCD(void){
	//LCD init, LCD module is preset for 8 bits data so we need to set it for 4 bits mode.
	delay(50);
	e=0; rs=0;
	PORTA=2; clock();	//4 bits mode
	PORTA=2; clock();	//function set: 4 bit
	PORTA=0; clock();	//function set: 1 line=0, 2 lines=8
	__delay_us(25); __delay_us(25);
	PORTA=0; clock();	//display set:
	PORTA=12; clock();	//display set: curzor off=12, curzor on=14, curzor on and flashing=15
	__delay_us(25); __delay_us(25);
	PORTA=0; clock(); PORTA=1; clock();	//display clear
	delay(2); 
	PORTA=0; clock();	//Entry mode:
	PORTA=6; clock();	//Entry mode:increment address	
}

void clock(void){		//LCD clock
	e=1;
	__delay_us(25);
	e=0;
	__delay_us(25);
}

void line(unsigned char msg){			//write a line to the lcd
	
//clear display
	rs=0;
	PORTA=0; clock(); PORTA=1; clock(); 	
	PORTA=0; clock(); PORTA=2; clock();	//return home
	delay(2);					//2ms
	
	switch (msg) {
		case 1: bk;bk;bk;bk;bk;
				digit((hour/10) %10+48);		//time
				digit(hour %10+48);
				digit(58);		//colon
				digit((min/10) %10+48);
				digit(min %10+48);
				
				break;
		case 2: digit(0x53);		//set hour
				digit(0x45);
				digit(0x54);bk;
				digit(0x48);
				digit(0x4F);
				digit(0x55);
				digit(0x52);
				bk;bk;
				digit((hour/10) %10+48);
				digit(hour %10+48);
				break;
		case 3: digit(0x53);		//set min
				digit(0x45);
				digit(0x54);bk;
				digit(0x4D);
				digit(0x49);
				digit(0x4E);
				bk;bk;
				digit(58);		//colon
				digit((min/10) %10+48);		//time
				digit(min %10+48);
				break;
		case 4: digit(0x52);		//record minutes tracks
				digit(0x45);
				digit(0x43);bk;
				digit(0x4D);
				digit(0x49);
				digit(0x4E);
				bk;
				digit((m_trk/10) %10+48);
				digit(m_trk %10+48);
				break;
		case 5: digit(0x53);		//SD err
				digit(0x44);bk;
				digit(0x65);
				digit(0x72);
				digit(0x72);
				break;
		case 6: digit(0x50);		//play minutes tracks
				digit(0x4C);
				digit(0x41);
				digit(0x59);
				bk;
				digit(0x4D);
				digit(0x49);
				digit(0x4E);
				bk;
				digit((m_trk/10) %10+48);
				digit(m_trk %10+48);
				break;
		case 7: bk;bk;digit(0x52);		//recording
				digit(0x45);
				digit(0x43);
				digit(0x4F);
				digit(0x52);
				digit(0x44);
				digit(0x49);
				digit(0x4E);
				digit(0x47);
				break;
		case 8: digit(0x52);		//record hour tracks
				digit(0x45);
				digit(0x43);bk;
				digit(0x48);
				digit(0x4F);
				digit(0x55);
				digit(0x52);
				bk;
				digit((h_trk/10) %10+48);
				digit(h_trk %10+48);
				break;
		case 9: digit(0x50);		//play hour tracks
				digit(0x4C);
				digit(0x41);
				digit(0x59);bk;
				digit(0x48);
				digit(0x4F);
				digit(0x55);
				digit(0x52);bk;
				digit((h_trk/10) %10+48);
				digit(h_trk %10+48);
				break;				
	}			
}

void digit(unsigned char fig){		//write a digit to the lcd

	rs=1;
	PORTA=(fig >> 4); clock();			//send 4 MSB
	PORTA=fig & 0b00001111; clock();	//send 4 LSB
	rs=0;

}

void initSD(void)		//initiate the SD
{
	unsigned char i;
	
	CS=1;
	for(i=0; i < 10; i++)SPI(0xFF);		// min 74 clocks
	CS=0;			// Enabled for SPI mode

	//if (Command(0x00,0,0x95) !=1) line(5);	// Reset SD command
	i=100;	//try reset for up to 100 times
	while(Command(0x00,0,0x95) !=1 && i!=0)
	{ 
		CS=1;
		SPI(0xFF);
		CS=0;
		i--;
	}
	if(i==0)	line(5);	//reset failed	
	
	while (Command(0x01,0,0xFF) !=0) {} //wait for idle state
	
	SSPM1 = 0;	// full speed 2MHz
	CS = 1;
}

unsigned char SPI(unsigned char data)		// send character over SPI
{
	SSPBUF = data;			// load character
	while (!BF);		// sent
	return SSPBUF;		// received character
}

char Command(unsigned char frame1, unsigned long adrs, unsigned char frame2 )
{	
	unsigned char i, res;
	SPI(0xFF);
	SPI((frame1 | 0x40) & 0x7F);	//first 2 bits are 01
	SPI((adrs & 0xFF000000) >> 24);		//first of the 4 bytes address
	SPI((adrs & 0x00FF0000) >> 16);
	SPI((adrs & 0x0000FF00) >> 8);
	SPI(adrs & 0x000000FF);	
	SPI(frame2 | 1);				//CRC and last bit 1

	for(i=0;i<10;i++)	// wait for received character
	{
		res = SPI(0xFF);
		if(res != 0xFF)break;
	}
	return res;	  
}

void writeSD(unsigned char mem)		//
{
	unsigned int r,i;
	unsigned char len=70;
	CS = 0;	
	ADON = 1;
	RecLED = 1;
	
	r = Command(25,51200000+mem*51200,0xFF);	//multi sector write
	if(r != 0)
	{
		line(5);
		ADON = 0;
		RecLED = 0;
	}
	SPI(0xFF);
	SPI(0xFF);
	SPI(0xFF);
	
	while(len)
	{
		SPI(0xFC);	//multi sector token byte
		for(i=0;i<512;i++)
		{
			#if defined __XC8
		    GO_DONE = 1;
		    #else
		    GODONE=1;
		    #endif
		    
			while(!TMR2IF){}	//20KHz clock
			SPI(ADRESL);		//send analogue byte
			TMR2IF = 0;
				//play while record
		//	CCP1X = ADRESL & 1;	//shift byte to get the required PWM duty cycle 
		//	CCPR1L = (ADRESL >> 1);
		}
 		SPI(0xFF);	// CRC
		SPI(0xFF);	// CRC
	
		if((r=SPI(0xFF) & 0x0F) == 0x05){	//data accepted	= 0101		
			for(i=10000;i>0;i--){				
				if(r=SPI(0xFF))	break;
			}	
		}
		else{
			line(5);
		}
		while(SPI(0xFF) != 0xFF);	// while busy
		--len;
	}
	SPI(0xFD);	//stop transfer	token byte
			
	SPI(0xFF);
	SPI(0xFF);
	while(SPI(0xFF) != 0xFF);	// while busy
	
	CS = 1;
	ADON = 0;
	RecLED = 0;
}

void readSD(unsigned char mem)
{
	unsigned int i,r;
	unsigned char data,len=70;

	CS = 0;
	r = Command(18,51200000+mem*51200,0xFF);	//read multi-sector
	if(r != 0)line(5);			//if command failed

	while(len)
	{
		while(SPI(0xFF) != 0xFE);	// wait for first byte
		for(i=0;i<512;i++){
			while(!TMR2IF){}
			data = SPI(0xFF);
			CCP1X = data & 1;	//shift byte to get the required PWM duty cycle 
			CCPR1L = (data >> 1);
			TMR2IF = 0;
		}
		SPI(0xFF);	//discard of CRC
		SPI(0xFF);
		--len;

	}
	
	Command(12,0x00,0xFF);	//stop transmit
	SPI(0xFF);
	SPI(0xFF);
	CS = 1;
}



